"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const __1 = require("../");
const ParserError_1 = require("../errors/ParserError");
const ParserErrors_1 = require("../errors/ParserErrors");
const Handler_1 = require("../lib/Handler");
const PathHelper_1 = require("../lib/PathHelper");
const HelperFunctions_1 = require("../lib/HelperFunctions");
class ParamParser {
    constructor() {
        this.params = [];
        this.onError = null;
    }
    withCustomErrorHandler(onError) {
        this.onError = onError;
        return this;
    }
    getAs(path, as, options) {
        return this.get(path, Object.assign({}, options, { as }));
    }
    get(path, options) {
        return this.add(Object.assign({}, options, { path }));
    }
    addAll(tmpOptions) {
        const options = _.isString(tmpOptions)
            ? JSON.parse(tmpOptions)
            : tmpOptions;
        Object.keys(options).forEach(key => this.add(Object.assign({}, options[key], { path: key })));
        return this;
    }
    add(options) {
        this.params.push(options);
        return this;
    }
    parse(data) {
        const params = PathHelper_1.PathHelper.reduceWildcardsInPaths(_.map(this.params, 'path'), (acc, _parentPath, expandedPath, index) => {
            acc.push(Object.assign({}, _.cloneDeep(this.params[index]), { path: expandedPath }));
            return acc;
        }, data, []);
        return this.parseParams(params, data);
    }
    parseParams(params, data) {
        return (params || []).reduce((acc, options) => {
            const { path, req, def, as, select, validate, sanitize, onError, } = options;
            const errorHandler = (value, errorCode) => {
                const errorId = ParserError_1.ParserErrorIds[errorCode];
                const parserError = ParserErrors_1.ParserErrorBuilders[errorId](path, value);
                _.isFunction(onError) && onError(path, value, parserError);
                _.isFunction(this.onError) && this.onError(path, value, parserError);
            };
            const defaultValue = _.isFunction(def)
                ? def()
                : def;
            const tmpValue = _.get(data, path);
            const value = tmpValue != null
                ? tmpValue
                : defaultValue;
            const required = req == null || req;
            if (value == null) {
                if (required) {
                    errorHandler(value, ParserError_1.ParserErrorCodes.RequiredPropNotSet);
                }
                return acc;
            }
            const sanitizedValue = Handler_1.Handler
                .reduceHandler(sanitize, (v, f) => __1.Sanitizers.invoke(f, `${v}`), value);
            const allValid = Handler_1.Handler
                .reduceHandler(validate, (v, f) => v && __1.Validators.invoke(f, sanitizedValue), true);
            if (!allValid) {
                errorHandler(sanitizedValue, ParserError_1.ParserErrorCodes.PropIsInvalid);
                return acc;
            }
            if (select != null && !select) {
                return acc;
            }
            const key = _.isString(as)
                ? as
                : _.last(path.split('.'));
            const mergeObj = HelperFunctions_1.safeInsertValueToPath(sanitizedValue, key);
            const merged = _.mergeWith(acc, mergeObj, (objValue, srcValue) => {
                if (_.isArray(objValue)) {
                    return objValue
                        .concat(srcValue)
                        .filter(x => x != null);
                }
                return undefined;
            });
            return merged;
        }, {});
    }
}
exports.ParamParser = ParamParser;
//# sourceMappingURL=ParamParser.js.map